@@ -3,14 +3,21 @@ import numpy as np
 import time
 import warnings
 
-from dadra.dyn_sys import DynamicSystem
-from graph_utils import plot_contour_2D, plot_contour_3D, plot_sample
-from sampling import num_samples
-from utils import (
-    compute_contour_2D,
-    compute_contour_3D,
-    empirical_estimate,
-    format_time,
+from dadra.dyn_sys import SimpleSystem, System
+from dadra.utils.graph_utils import plot_contour_2D, plot_contour_3D, plot_sample
+from dadra.utils.christoffel_utils import (
+    c_compute_contours,
+    c_emp_estimate,
+    c_num_samples,
+    construct_inv_christoffel,
+    construct_kernelized_inv_christoffel,
+)
+from dadra.utils.misc_utils import format_time
+from dadra.utils.p_utils import (
+    p_compute_contour_2D,
+    p_compute_contour_3D,
+    p_emp_estimate,
+    p_num_samples,
     solve_p_norm,
 )
 
@@ -18,28 +25,46 @@ from utils import (
 class Estimator:
     """Class which allows for estimation of the reachable sets for a given dynamical system.
 
-    :param dyn_sys: An instance of :class:`dadra.DynamicSystem`
-    :type dyn_sys: DynamicSystem
+    :param dyn_sys: An instance of :class:`dadra.System`
+    :type dyn_sys: System
     :param epsilon: The accuracy parameter, defaults to 0.05
     :type epsilon: float, optional
     :param delta: The confidence parameter, defaults to 1e-9
     :type delta: float, optional
+    :param christoffel: If True, uses Christoffel functions to estimate the reachable set. If False, the p-norm method is used, defaults to False
+    :type christoffel: bool, optional
     :param p: The order of p-norm, defaults to 2
     :type p: int, optional
     :param const: The constraints placed on the parameters A and b, defaults to None
     :type const: string, optional
     :param normalize: If true, the sample is normalized, defaults to True
     :type normalize: bool, optional
+    :param d: The degree of polynomial features points are mapped to, defaults to 10
+    :type d: int, optional
+    :param kernelized: If True, the kernelized inverse christoffel function is constructed, defaults to False
+    :type kernelized: bool, optional
+    :param rho: Inverse Christoffel function constant rho, defaults to 0.0001
+    :type rho: float, optional
+    :param rbf: If True, a radical basis function kernel is used, defaults to False
+    :type rbf: bool, optional
+    :param scale: The length scale of the kernel, defaults to 1.0
+    :type scale: float, optional
     """
 
     def __init__(
         self,
-        dyn_sys: DynamicSystem,
+        dyn_sys: System,
         epsilon=0.05,
         delta=1e-9,
+        christoffel=False,
         p=2,
         const=None,
         normalize=True,
+        d=10,
+        kernelized=False,
+        rho=0.0001,
+        rbf=False,
+        scale=1.0,
     ):
 
         self.dyn_sys = dyn_sys
@@ -47,13 +72,31 @@ class Estimator:
 
         self.epsilon = epsilon
         self.delta = delta
-        self.p = p
-        self.const = const
+        self.christoffel = christoffel
+
+        if not christoffel:
+            self.p = p
+            self.const = const
+        else:
+            self.const = const
+            self.d = d
+            self.kernelized = kernelized
+            self.rho = rho
+            self.rbf = rbf
+            self.scale = scale
+
         self.normalize = normalize
         self.num_samples = self.get_num_samples()
-        self.samples = self.get_sample()
 
-        self.A, self.b, self.status = self.solve_p()
+        self.samples = None
+
+        if not christoffel:
+            self.A = None
+            self.b = None
+            self.status = None
+        else:
+            self.C = None
+            self.level = None
 
     @classmethod
     def estimator_from_func(
@@ -65,13 +108,19 @@ class Estimator:
         parts=1001,
         epsilon=0.05,
         delta=1e-9,
+        christoffel=False,
         p=2,
         const=None,
         normalize=True,
+        d=10,
+        kernelized=False,
+        rho=0.0001,
+        rbf=False,
+        scale=0.5,
     ):
         """Class method that allows for an instance of :class:`dadra.Estimator` to be initialized
-        using a dynamic function, and the components for an instance of :class:`DynamicSystem`
-        rather than explicitly passing in a :class:`DynamicSystem` object
+        using a dynamic function, and the components for an instance of :class:`SimpleSystem`
+        rather than explicitly passing in a :class:`System` object
 
         :param dyn_func: The function defining the dynamics of the system
         :type dyn_func: function
@@ -87,17 +136,42 @@ class Estimator:
         :type epsilon: float, optional
         :param delta: The confidence parameter, defaults to 1e-9
         :type delta: float, optional
+        :param christoffel: If True, uses Christoffel functions to estimate the reachable set. If False, the p-norm method is used, defaults to False
+        :type christoffel: bool, optional
         :param p: The order of p-norm, defaults to 2
         :type p: int, optional
         :param const: The constraints placed on the parameters A and b, defaults to None
         :type const: string, optional
         :param normalize: If true, the sample is normalized, defaults to True
         :type normalize: bool, optional
+        :param d: The degree of polynomial features points are mapped to, defaults to 10
+        :type d: int, optional
+        :param kernelized: If True, the kernelized inverse christoffel function is constructed, defaults to False
+        :type kernelized: bool, optional
+        :param rho: Inverse Christoffel function constant rho, defaults to 0.0001
+        :type rho: float, optional
+        :param rbf: If True, a radical basis function kernel is used, defaults to False
+        :type rbf: bool, optional
+        :param scale: The length scale of the kernel, defaults to 1.0
+        :type scale: float, optional
         :return: A :class:`dadra.Estimator` object
         :rtype: :class:`dadra.Estimator`
         """
-        dyn_sys = DynamicSystem(dyn_func, intervals, state_dim, timesteps, parts)
-        return cls(dyn_sys, epsilon, delta, p, const, normalize)
+        dyn_sys = SimpleSystem(dyn_func, intervals, state_dim, timesteps, parts)
+        return cls(
+            dyn_sys,
+            epsilon=epsilon,
+            delta=delta,
+            christoffel=christoffel,
+            p=p,
+            const=const,
+            normalize=normalize,
+            d=d,
+            kernelized=kernelized,
+            rho=rho,
+            rbf=rbf,
+            scale=scale,
+        )
 
     @classmethod
     def estimator_from_list(
@@ -108,13 +182,19 @@ class Estimator:
         parts=1001,
         epsilon=0.05,
         delta=1e-9,
+        christoffel=False,
         p=2,
         const=None,
         normalize=True,
+        d=10,
+        kernelized=False,
+        rho=0.0001,
+        rbf=False,
+        scale=0.5,
     ):
         """Class method that allows for an instance of :class:`dadra.Estimator` to be initialized
-        using a list of dynamic functions, and the components for an instance of :class:`DynamicSystem`
-        rather than explicitly passing in a :class:`DynamicSystem` object
+        using a list of dynamic functions, and the components for an instance of :class:`SimpleSystem`
+        rather than explicitly passing in a :class:`System` object
 
         :param dyn_func: The list of functions, one for each variable, that define the dynamics of the system
         :type dyn_func: function
@@ -128,25 +208,50 @@ class Estimator:
         :type epsilon: float, optional
         :param delta: The confidence parameter, defaults to 1e-9
         :type delta: float, optional
+        :param christoffel: If True, uses Christoffel functions to estimate the reachable set. If False, the p-norm method is used, defaults to False
+        :type christoffel: bool, optional
         :param p: The order of p-norm, defaults to 2
         :type p: int, optional
         :param const: The constraints placed on the parameters A and b, defaults to None
         :type const: string, optional
         :param normalize: If true, the sample is normalized, defaults to True
         :type normalize: bool, optional
+        :param d: The degree of polynomial features points are mapped to, defaults to 10
+        :type d: int, optional
+        :param kernelized: If True, the kernelized inverse christoffel function is constructed, defaults to False
+        :type kernelized: bool, optional
+        :param rho: Inverse Christoffel function constant rho, defaults to 0.0001
+        :type rho: float, optional
+        :param rbf: If True, a radical basis function kernel is used, defaults to False
+        :type rbf: bool, optional
+        :param scale: The length scale of the kernel, defaults to 1.0
+        :type scale: float, optional
         :return: A :class:`dadra.Estimator` object
         :rtype: :class:`dadra.Estimator`
         """
-        dyn_sys = DynamicSystem.get_system(func_list, intervals, timesteps, parts)
-        return cls(dyn_sys, epsilon, delta, p, const, normalize)
+        dyn_sys = SimpleSystem.get_system(func_list, intervals, timesteps, parts)
+        return cls(
+            dyn_sys,
+            epsilon=epsilon,
+            delta=delta,
+            christoffel=christoffel,
+            p=p,
+            const=const,
+            normalize=normalize,
+            d=d,
+            kernelized=kernelized,
+            rho=rho,
+            rbf=rbf,
+            scale=scale,
+        )
 
     def check_sys(self):
         """Checks whether the dynamic system passed to the constructor is valid
 
-        :raises TypeError: If an object of type other than :class:`dadra.DynamicSystem` is passed
+        :raises TypeError: If an object of type other than :class:`dadra.System` is passed
         """
-        if not isinstance(self.dyn_sys, DynamicSystem):
-            raise TypeError("Object of type DynamicSystem must be passed")
+        if not isinstance(self.dyn_sys, System):
+            raise TypeError("Object of type System must be passed")
 
     def get_num_samples(self):
         """Compute the number of samples needed to satisfy the specified probabilistic guarantees
@@ -155,7 +260,10 @@ class Estimator:
         :rtype: int
         """
         n_x = self.dyn_sys.state_dim
-        return num_samples(self.epsilon, self.delta, n_x, self.const)
+        if not self.christoffel:
+            return p_num_samples(self.epsilon, self.delta, n_x, self.const)
+        else:
+            return c_num_samples(self.epsilon, self.delta, n_x, self.d)
 
     def get_sample(self):
         """Draws the number of samples necessary to satisfy the specified probabilistic guarantees
@@ -176,6 +284,15 @@ class Estimator:
             samples = (samples - np.mean(samples, axis=0)) / np.std(samples)
         return samples
 
+    def estimate(self):
+        """Draws the samples and makes an estimate of the reachable set"""
+        self.samples = self.get_sample()
+
+        if not self.christoffel:
+            self.A, self.b, self.status = self.solve_p()
+        else:
+            self.C, self.level = self.construct_christoffel()
+
     def solve_p(self):
         """Solves for the optimal p-norm ball that estimates the reachable set
 
@@ -196,13 +313,30 @@ class Estimator:
             )
         return A, b, status
 
+    def construct_christoffel(self):
+        if self.kernelized:
+            C = construct_kernelized_inv_christoffel(
+                self.samples, self.d, rho=self.rho, rbf=self.rbf, scale=self.scale
+            )
+        else:
+            C = construct_inv_christoffel(self.samples, self.d)
+
+        start_lvl_param = time.perf_counter()
+        values = C(self.samples)
+        level = np.max(values)
+        end_lvl_param = time.perf_counter()
+        print(
+            f"Time to compute level parameter: {format_time(end_lvl_param - start_lvl_param)}"
+        )
+        return C, level
+
     def empirical_estimate(self, num_samples_emp=None):
         """Computes the ratio of samples within the estimated reachable set
 
         :param num_samples_emp: The number of samples to draw for the empirical estimate, defaults to the number of samples drawn to obtain the p-norm ball
         :type num_samples_emp: int, optional
         """
-        if self.status != "optimal":
+        if not self.christoffel and self.status != "optimal":
             warnings.warn(
                 "Non-optimal solution to p-norm ball. Empirical estimate may be innaccurate"
             )
@@ -224,10 +358,14 @@ class Estimator:
                 samples_emp
             )
 
-        ratio = empirical_estimate(
-            samples_emp, self.A, self.b, self.dyn_sys.state_dim, self.p
-        )
-        print(f"Ratio of samples within the estimated reachability set: {ratio}")
+        if not self.christoffel:
+            ratio = p_emp_estimate(
+                samples_emp, self.A, self.b, self.dyn_sys.state_dim, self.p
+            )
+        else:
+            ratio = c_emp_estimate(self.samples, self.C, self.level)
+
+        print(f"Ratio of samples within the estimated reachable set: {ratio}")
 
     def summary(self):
         """Prints a summary of the attributes of this instance of :class:`Estimator`"""
@@ -241,10 +379,35 @@ class Estimator:
         summary_str += f"State dimension: {self.dyn_sys.state_dim}" + "\n"
         summary_str += f"Accuracy parameter epsilon: {self.epsilon}" + "\n"
         summary_str += f"Confidence parameter delta: {self.delta}" + "\n"
-        summary_str += f"p-norm p value: {self.p}" + "\n"
-        summary_str += f"Constraints on p-norm ball: {self.const}" + "\n"
         summary_str += f"Number of samples: {self.num_samples}" + "\n"
-        summary_str += f"Status of p-norm ball solution: {self.status}" + "\n"
+        if not self.christoffel:
+            summary_str += f"Method of estimation: p-Norm Ball"
+            summary_str += f"p-norm p value: {self.p}" + "\n"
+            summary_str += f"Constraints on p-norm ball: {self.const}" + "\n"
+            if all([self.A is None, self.b is None, self.status is None]):
+                summary_str += (
+                    f"Status of p-norm ball solution: No estimate has been made yet"
+                    + "\n"
+                )
+            else:
+                summary_str += f"Status of p-norm ball solution: {self.status}" + "\n"
+        else:
+            summary_str += f"Method of estimation: Inverse Christoffel Function" + "\n"
+            summary_str += f"Degree of polynomial features: {self.d}" + "\n"
+            summary_str += f"Kernelized: {self.kernelized}" + "\n"
+            summary_str += f"Constant rho: {self.rho}" + "\n"
+            summary_str += f"Radical basis function kernel: {self.rbf}" + "\n"
+            summary_str += f"Length scale of kernel: {self.scale}" + "\n"
+            if all([self.C is None, self.level is None]):
+                summary_str += (
+                    f"Status of Christoffel Function estimate: No estimate has been made yet"
+                    + "\n"
+                )
+            else:
+                summary_str += (
+                    f"Status of Christoffel Function estimate: Estimate made" + "\n"
+                )
+
         summary_str += (
             "----------------------------------------------------------------" + "\n"
         )
@@ -266,40 +429,51 @@ class Estimator:
         :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
         :type grid_n: int
         """
-        cont_compute = partial(
-            compute_contour_2D,
-            sample=self.samples,
-            A_val=self.A,
-            b_val=self.b,
-            n_x=self.dyn_sys.state_dim,
-            p=self.p,
-            grid_n=grid_n,
-        )
+        if not self.christoffel:
+            cont_compute = partial(
+                p_compute_contour_2D,
+                sample=self.samples,
+                A_val=self.A,
+                b_val=self.b,
+                n_x=self.dyn_sys.state_dim,
+                p=self.p,
+                grid_n=grid_n,
+            )
 
-        print("Computing 2D contours")
-        cont_start = time.perf_counter()
-        xv1, yv1, z_cont, z_min, z_max = cont_compute(cont_axis=2)
-        xv2, zv1, y_cont, y_min, y_max = cont_compute(cont_axis=1)
-        yv2, zv2, x_cont, x_min, x_max = cont_compute(cont_axis=0)
-        cont_end = time.perf_counter()
-        print(f"Time to compute 2D contours: {format_time(cont_end - cont_start)}")
+            print("Computing 2D contours")
+            cont_start = time.perf_counter()
+            xv1, yv1, z_cont, z_min, z_max = cont_compute(cont_axis=2)
+            xv2, zv1, y_cont, y_min, y_max = cont_compute(cont_axis=1)
+            yv2, zv2, x_cont, x_min, x_max = cont_compute(cont_axis=0)
+            cont_end = time.perf_counter()
+            print(f"Time to compute 2D contours: {format_time(cont_end - cont_start)}")
 
-        plot_contour_2D(
-            xv1,
-            yv1,
-            z_cont,
-            z_max,
-            self.samples,
-            fig_name,
-            xv2=xv2,
-            zv1=zv1,
-            y_cont=y_cont,
-            y_level=y_max,
-            yv2=yv2,
-            zv2=zv2,
-            x_cont=x_cont,
-            x_level=x_max,
-        )
+            plot_contour_2D(
+                xv1,
+                yv1,
+                z_cont,
+                z_max,
+                self.samples,
+                fig_name,
+                xv2=xv2,
+                zv1=zv1,
+                y_cont=y_cont,
+                y_level=y_max,
+                yv2=yv2,
+                zv2=zv2,
+                x_cont=x_cont,
+                x_level=x_max,
+            )
+
+        else:
+            start_comp_cont = time.perf_counter()
+            xv, yv, cont = c_compute_contours(self.C, self.samples, grid_n)
+            end_comp_cont = time.perf_counter()
+            print(
+                f"Time to compute contour: {format_time(end_comp_cont - start_comp_cont)}"
+            )
+
+            plot_contour_2D(xv, yv, cont, self.level, self.samples, fig_name)
 
     def plot_3D_cont(self, fig_name, grid_n=100, gif_name=None):
         """Computes and plots the contours in 3D with the option for saving an animated gif of the rotating graph
@@ -313,7 +487,7 @@ class Estimator:
         """
         print("Computing 3D contour")
         cont_start = time.perf_counter()
-        d0, d1, cont_min, cont_max, c_min, c_max = compute_contour_3D(
+        d0, d1, cont_min, cont_max, c_min, c_max = p_compute_contour_3D(
             self.samples, self.A, self.b, grid_n=grid_n
         )
         cont_end = time.perf_counter()
